close all; clear; clc;
addpath(genpath('.\utils'))

%
% DEPcode (2018):
% This code implements a version of the descriptor described in the paper:
%
% Discrete time Evolution Process Descriptor for shape analysis and matching
% ACM Transactions on Graphics, Vol. 37, No. 1, Article 4. January 2018
%
% Simone Melzi - University of Verona
% Maks Ovsjanikov - LIX cole Polytechnique 
% Giorgio Roffo - University of Glasgow
% Marco Cristani - University of Verona
% Umberto Castellani - University of Verona
%
% use DEP operator in the functional maps framework similarly to figures 
% 11,12 in the paper
%
% original parameters;
% c = 0.5; flag_dist = 'Dijkstra'; numEigsSrc = 60; numEigsTar = 60; sym = 1;
%
% If you set flag_dist = 'SpectroMeter' it is 10 times faster.
%

%% load shapes
[X,T] = readOff('.\data\tr_reg_089');
Src = MeshInfo(X, T, 200);

[X,T] = readOff('.\data\tr_reg_090');
Tar = MeshInfo(X, T, 200);

%% select parameters
% DEP parameters
c = 0.5; % 0.1; % 0.9;
flag_dist = 'SpectroMeter'; % 'Dijkstra'; % SpectroMeter is 10 times faster

% fmaps parameters
numEigsSrc = 60;
numEigsTar = 60;
sym = 1;

%% prepare landmarks and descriptors
% WKS
Src.probe = [];
Tar.probe = [];
Src.probe = [Src.probe, waveKernelSignature(Src.laplaceBasis, Src.eigenvalues, Src.Ae, 100)]; % n_scale);
Tar.probe = [Tar.probe, waveKernelSignature(Tar.laplaceBasis, Tar.eigenvalues, Tar.Ae, 100)]; %n_scale);
Src.probe = [Src.probe, heatKernelSignature(Src, 100)]; % n_scale);
Tar.probe = [Tar.probe, heatKernelSignature(Tar, 100)]; %n_scale);

% DEP
disp('compute distances'); tic;
if strcmp(flag_dist,'Dijkstra')
    Src.dists = zeros(Src.nv);
    shape.X = Src.X(:,1); shape.Y = Src.X(:,2); shape.Z = Src.X(:,3); shape.TRIV = Src.T;
    for j = 1:Src.nv
        Src.dists(j,:) = dijkstra_to_all(shape,j);        
    end
    clear shape;
    Tar.dists = zeros(Tar.nv);
    shape.X = Tar.X(:,1); shape.Y = Tar.X(:,2); shape.Z = Tar.X(:,3); shape.TRIV = Tar.T;
    for j = 1:Tar.nv
        Tar.dists(j,:) = dijkstra_to_all(shape,j);        
    end
elseif strcmp(flag_dist,'SpectroMeter')
    [Src.dists] = SpectroMeterDists(Src); 
    [Tar.dists] = SpectroMeterDists(Tar);     
end
toc

% compute diameter and normalize distanes
Src.dists = Src.dists./max(Src.dists(:));
Tar.dists = Tar.dists./max(Tar.dists(:));

% compute area elements
Src.Omega = abs(diag(sum(full(Src.Ae),2)));
Tar.Omega = abs(diag(sum(full(Tar.Ae),2)));
Src.D_norm = Src.dists;
Src.D_norm = ones(size(Src.D_norm)) - Src.D_norm; 
Tar.D_norm = Tar.dists;
Tar.D_norm = ones(size(Tar.D_norm)) - Tar.D_norm; 

% compute DEP descriptor
tic
disp('compute DEP operator')
Src.DEPop = computeDEPop(Src.D_norm*Src.Omega,c);
Tar.DEPop = computeDEPop(Tar.D_norm*Tar.Omega,c);
toc

%% compute the fmaps
C_gt = Tar.laplaceBasis(:,1:numEigsTar)'*Tar.Ae*Src.laplaceBasis(:,1:numEigsSrc);
% LBO
Src.operators{1} = Src.cotLaplacian; Tar.operators{1} = Tar.cotLaplacian; 
C_LBO = compute_map_operators(Src,Tar,numEigsSrc); 
% DEP
Src.operators{1} = Src.DEPop; Tar.operators{1} = Tar.DEPop; 
C_DEP = compute_map_operators(Src,Tar,numEigsSrc);
% LBO+DEP
Src.operators{1} = Src.cotLaplacian; Tar.operators{1} = Tar.cotLaplacian; 
Src.operators{2} = Src.DEPop; Tar.operators{2} = Tar.DEPop; 
C_LBODEP = compute_map_operators(Src,Tar,numEigsSrc); 

%% compute the p2p maps
p2p_gt = knnsearch(Tar.laplaceBasis(:,1:numEigsTar),( C_gt*(Src.laplaceBasis(:,1:numEigsSrc)'))');
p2p_LBO = knnsearch(Tar.laplaceBasis(:,1:numEigsTar),( C_LBO*(Src.laplaceBasis(:,1:numEigsSrc)'))');
p2p_DEP = knnsearch(Tar.laplaceBasis(:,1:numEigsTar),( C_DEP*(Src.laplaceBasis(:,1:numEigsSrc)'))');
p2p_LBODEP = knnsearch(Tar.laplaceBasis(:,1:numEigsTar),( C_LBODEP*(Src.laplaceBasis(:,1:numEigsSrc)'))');

%%
max_th = 0.3; thresh = 0:0.001:1;
D = Tar.dists; diam = max(D(:));
idx = find(thresh==max_th);
errs = zeros(Src.nv,4);

for i=1:Src.nv
    if sym == 1
        tmp = load(['.\data\sym_labels_FAUST.mat']); 
        sym = tmp.idxs; clear tmp;
        errs(i,1) = min(D(i,p2p_gt(i)),D(sym(i),p2p_gt(i)));
        errs(i,2) = min(D(i,p2p_LBO(i)),D(sym(i),p2p_LBO(i)));
        errs(i,3) = min(D(i,p2p_DEP(i)),D(sym(i),p2p_DEP(i)));
        errs(i,4) = min(D(i,p2p_LBODEP(i)),D(sym(i),p2p_LBODEP(i)));
        
    else
        errs(i,1) = D(i,p2p_gt(i));
        errs(i,2) = D(i,p2p_LBO(i));
        errs(i,3) = D(i,p2p_DEP(i));
        errs(i,4) = D(i,p2p_LBODEP(i));
    end
end
errs = errs ./ diam;
curve_gt = calc_err_curve(errs(:,1), thresh);
curve_LBO = calc_err_curve(errs(:,2), thresh);
curve_DEP = calc_err_curve(errs(:,3), thresh);
curve_LBODEP = calc_err_curve(errs(:,4), thresh);
AUCs(1) = trapz(thresh(1:idx), curve_gt(1:idx))/max_th; 
AUCs(2) = trapz(thresh(1:idx), curve_LBO(1:idx))/max_th; 
AUCs(3) = trapz(thresh(1:idx), curve_DEP(1:idx))/max_th; 
AUCs(4) = trapz(thresh(1:idx), curve_LBODEP(1:idx))/max_th; 

figure
plot(thresh, curve_gt, 'linewidth', 2); hold on;
plot(thresh, curve_LBO, 'linewidth', 2);
plot(thresh, curve_DEP, 'linewidth', 2);
plot(thresh, curve_LBODEP, 'linewidth', 2);
set(gca, 'xlim', [0 max_th])
set(gca, 'ylim', [0 100])
set(gca, 'xgrid', 'on')
set(gca, 'ygrid', 'on')
xlabel('Geodesic error')
ylabel('% Correspondences')
title('p2p evaluation')
legend(['gt ', num2str(AUCs(1))],['LBO ', num2str(AUCs(2))],['DEP ', num2str(AUCs(3))],['LBODEP ', num2str(AUCs(4))],'Location','southeast');
